
#include <inc/string.h>
#include <inc/lib.h>

void
cputchar(int ch)
{
	char c = ch;

	// Unlike standard Unix's putchar,
	// the cputchar function _always_ outputs to the system console.
	sys_cputs(&c, 1);
}

int
getchar(void)
{
	unsigned char c;
	int r;

	// JOS does, however, support standard _input_ redirection,
	// allowing the user to redirect script files to the shell and such.
	// getchar() reads a character from file descriptor 0.
	r = read(0, &c, 1);
	if (r < 0)
		return r;
	if (r < 1)
		return -E_EOF;
	return c;
}


// "Real" console file descriptor implementation.
// The putchar/getchar functions above will still come here by default,
// but now can be redirected to files, pipes, etc., via the fd layer.

struct ConsoleDev : public Dev {
	ConsoleDev() : Dev('c', "cons") {
	}
	
	ssize_t read(Fd *fd, void *buf, size_t n, off_t offset);
	ssize_t write(Fd *fd, const void *buf, size_t n, off_t offset);
	int close(Fd *fd);
	int stat(Fd *fd, struct Stat *stat);
};

ConsoleDev console_dev;

int
iscons(int fdnum)
{
	Fd *fd = Fd::fd(fdnum, true);
	return fd && fd->device() == &console_dev;
}

int
opencons(void)
{
	int r;
	struct Fd* fd;

	if (!(fd = Fd::unused_fd()))
		return -E_MAX_OPEN;
	if ((r = sys_page_alloc(0, fd, PTE_P|PTE_U|PTE_W|PTE_SHARE)) < 0)
		return r;
	fd->fd_type = console_dev.dev_id;
	fd->fd_omode = O_RDWR;
	return fd->fd_number();
}

ssize_t
ConsoleDev::read(struct Fd *fd, void *vbuf, size_t n, off_t offset)
{
	int c;

	USED(offset);

	if (n == 0)
		return 0;

	while ((c = sys_cgetc()) == 0)
		sys_yield();
	if (c < 0)
		return c;
	if (c == 0x04)	// ctl-d is eof
		return 0;
	*(char*)vbuf = c;
	return 1;
}

ssize_t
ConsoleDev::write(struct Fd *fd, const void *vbuf, size_t n, off_t offset)
{
	unsigned tot, m;
	char buf[128];

	USED(offset);

	// mistake: have to nul-terminate arg to sys_cputs, 
	// so we have to copy vbuf into buf in chunks and nul-terminate.
	for (tot = 0; tot < n; tot += m) {
		m = n - tot;
		if (m > sizeof(buf) - 1)
			m = sizeof(buf) - 1;
		memcpy(buf, (char*)vbuf + tot, m);
		sys_cputs(buf, m);
	}
	return tot;
}

int
ConsoleDev::close(struct Fd *fd)
{
	USED(fd);

	return 0;
}

int
ConsoleDev::stat(struct Fd *fd, struct Stat *stat)
{
	strcpy(stat->st_name, "<cons>");
	return 0;
}

